'
'	YAOSM version 1.7
'

#include "12F629_675.h"		' 12F629 and 12F675
'#include "12F635.h"			' 12F635
'#include "12F683.h"			' 12F683
'#include "16F627_628.h"		' 16F627, 16F628, 16F627A, 16F628A and 16F648A
'#include "16F630_676.h"		' 16F630 and 16F676
'#include "16F636_639.h"		' 16F636 and 16F639
'#include "16F684.h"			' 16F684

'#define dDebug				' Enable debugging code
#define dGCRegionPatching	' Enable GC region patching
#define dWiiRegionPatching	' Enable Wii region patching
#define dMediaSupport		' Enable DVD-RW, DVD+R and DVD+RW support
'#define dFlashLED			' Flash LED2 while waiting for disc
#define dEpromFlash			' Flash LED1 four times after writing to eeprom
#define dConfigDisc			' Enable support for config disc

	movlw	dDirections		' Saving some space by setting up all I/O directions at the same time
	movwf	dTrisio

	set pLed1 on	' LED1 indicates start of first sync
	bUnlock = 0		' Clear bUnlock to indicate that chip is undetected and sync will therefor not unlock chip
	sSyncAndUnlock	' Sync with cn302 (no unlock)
	set pLed1 off	' turn off LED1 to indicate sync success

	dim wNewstack as word									' wNewstack is a word
	bCmd  = 0xFF
	bHigh = 0x40
	bMid  = 0xBC
	bLow  = 0xB2
	bNum  = 1
	bFirstbyte = 0
	bSecondbyte = 0
	sWriteCommand				' Read 0x40BCB2 to check for D2B chip (0x40BCB2 = 0x20 on D2B)
	if bResponse = 0x20 then	' D2B found
		bUnlock = 0xBE										' Setup variables
		bMedia 	= 0x80
		bSpeed	= 0x86
		wNewstack=0xD53E ' d54a - 0x0c
	end if
	if bUnlock = 0 then										' Assume DMS/D2A if D2B is not found
		bUnlock = 0xB6										' Setup variables
		bMedia 	= 0x76
		bSpeed	= 0x7A
		wNewstack=0xD4BB ' d4c7 - 0x0c
	end if

	sSyncAndUnlock				' Sync and unlock now that the unlock address is known

ReadSettings:
	EPRead(0, bMyregion)		' Read region setting from eprom (default is 0xFF = disabled)
	EPRead(1, bMyspeed1)		' Read Wii speed setting from eprom  (default is 0xFF = Wii speed)
	EPRead(2, bMyspeed2)		' Read alternate Wii speed setting from eprom  (default is 0xFF = Wii speed)
	
	#ifdef dWiiRegionPatching					' Setup regionpatching (if enabled)
		bRegionlow = 0xD6						' JAP	
		if bMyregion = 1 then bRegionlow = 0xD7	' USA
		if bMyregion = 2 then bRegionlow = 0xDA	' EUR
	#endif


AllDone:				' Start of main loop, reset leds and clear some variables
	set pLed1 off		' Turn off LED1 since we are not patching at the moment
	set pLed2 on		' Turn on LED2 to show that we are waiting for a disc
	#ifdef dDebug
		bTemp = 0		' Reset counter, used for led flashing when debugging
	#endif

	bHigh = 0x40

	if bBackup = 6 then									' Only patch speed for backups
		if bGamecube > 0 then goto SkipSpeedPatching	' Never ever change speed for Gamecube games
		if bMyspeed = 0 then goto SkipSpeedPatching		' Are we supposed to patch?
		bCmd  = 0xFE
'		bHigh = 0x40
		bMid  = 0xBD
		bLow  = bSpeed
		bNum  = 2
'		bSecondbyte = 0
		if bMyspeed = 1 then bFirstbyte = 0x81			' Medium speed(0x0081) (somewhere between?)
		if bMyspeed > 1 then 							' Fast speed  (0x01BD) (wii speed)
			bFirstbyte = 0xBD
			bSecondbyte = 0x01
		end if
		sWriteCommand									' Apply the new drive speed
	end if

SkipSpeedPatching:
	bBackup = 0			' Clear backup flag (doubles as region offset)

WaitForDisc:
	#ifdef dFlashLED						' Flash LED2 while waiting for disc when debug is enabled
		if bTemp.7 = 0 then set pLed2 on	' 0-127
		if bTemp.7 = 1 then set pLed2 off	' 128-255
		bTemp = bTemp + 1					' Increase counter
	#endif
	bCmd  = 0xFF							' Read login state
	bHigh = 0x40
	bMid  = 0xBA
	bLow  = 0x06
	bNum  = 1
	bFirstbyte = 0
	bSecondbyte = 0
	sWriteCommand
	if bResponse = 0x51 then goto DetectedDisc	' Go to next step if disc detected
	#ifdef dConfigDisc
		if bGamecube = 2 then						' Only run when we have the correct disc in the drive
			bCmd  = 0xFF							' Read DI buffer
			bHigh = 0x00
			bMid  = 0x83
			bLow  = bUnlock + 0x52
			bNum  = 2
'			bFirstbyte = 0
'			bSecondbyte = 0
			sWriteCommand
			if bResponse = 0xAA then				' 0xAA is our faked DI-command
				if bResponse2 = 0xAA then			' Are we setting a value?
					bLow = bUnlock + 0x54
					sWriteCommand					' Read DI buffer + 2 (pos + value)
					EPWrite(bResponse,bResponse2)	' Save in eprom
					bCmd = 0xFE
					bLow = bUnlock + 0x52
					bFirstbyte = 0xAA
					sWriteCommand					' Blank second byte so that we do not set it again
				end if
				if bResponse2 = 0xAB then 			' Reinitialize configuration?
					bCmd = 0xFE
					bFirstbyte = 0xAA
					sWriteCommand					' Blank second byte so that we do not do this again
					#ifdef dEpromFlash
						sQuadrupleFlash				' Flash LED1 four times
					#endif
					goto ReadSettings				' Read settings from eprom if we are done
				end if
			end if
		end if
	#endif
	goto WaitForDisc							' Loop until disc state is 0x51
	
DetectedDisc:
	set pLed2 off								' Turn off LED2 to show that we found a disc
	
'	bCmd  = 0xFF
'	bHigh = 0x40
'	bMid  = 0xBA
'	bLow  = 0x06
'	bNum  = 1
'	bFirstbyte = 0
'	bSecondbyte = 0
WaitFor7F:
	sWriteCommand								' Read login state
	if bResponse < 0x51 then goto AllDone		' Avoid getting stuck in loop, exit if loginstate drops below 0x51
	if bResponse <> 0x7F then goto WaitFor7F	' Wait for 0x7F

Detected7F:	
	bCmd  = 0xFE					' Write new return address to stack (WiiFree 2.31+ way of doing it) (Buy some time?)
	bHigh = 0x00
	bMid  = 0x8F
	bLow  = 0x84
	bNum  = 2
	bFirstbyte = wNewstack
	bSecondbyte = wNewstack_h
	sWriteCommand	

	bCmd  = 0xFF
	bNum  = 1
	bFirstbyte = 0
	bSecondbyte = 0
WaitForMedia:
	bHigh = 0x00
	bMid  = 0x85
	bLow  = bMedia
	sWriteCommand												' Read mediabyte
	if bResponse = 0x18 then goto DetectedBackup				' Check if backup media
	if bResponse = 0x0c then goto DetectedBackup				' Check if backup media
'	bCmd  = 0xFF												' Read login state
	bHigh = 0x40
	bMid  = 0xBA
	bLow  = 0x06
'	bNum  = 1
'	bFirstbyte = 0
'	bSecondbyte = 0
	sWriteCommand												
	if bResponse < 0x51 then goto AllDone						' Avoid getting stuck in loop, exit if loginstate drops below 0x51
	if bResponse = 0xFF then goto WaitForMedia					' In rare occations login state returns 0xFF and then we read it again
	if bResponse > 0xEF then goto BackupPatchingDone			' If we reach login state 0xF0 or above here then it's an original
	goto WaitForMedia											' Loop
	
DetectedBackup:	
	set pLed1 on												' LED1 on means backup patching has started
   	' Patch so that backup will be accepted
	bCmd  = 0xFE
'	bHigh = 0x00
'	bMid  = 0x85
'	bLow  = bMedia
'	bNum  = 1
	bFirstbyte = 0x28
'	bSecondbyte = 0
	sWriteCommand		' Write 0x28 to media
	bLow  = bLow + 0x22
	bFirstbyte = 0x04
	sWriteCommand		' Write 0x04 to emu
	#ifdef dMediaSupport
		' Patch media for unsupported media (DVD-RW, DVD+R, DVD+RW) (if this takes to long, then GC backups can fail with black screen and error)
		bCmd  = 0xFF				' Read valid number of sectors (highest and high)
		bHigh = 0x40
		bMid  = 0x88
		bLow  = 0x14
		bNum  = 2
		bFirstbyte = 0
'		bSecondbyte = 0
		sWriteCommand				
'		bTemp2		= bResponse		' highest byte (should be zero since largest possible sector count is 4,173,824 (DVD+DL))
		bTemp	    = bResponse2	' high byte
	
		bLow  = 0x16				' Read valid number of sectors (mid and low)
		sWriteCommand				
		
		bCmd  = 0xFE				' Write the valid number of sectors (mid and low)
		bHigh = 0x00
		bMid  = 0x85
		bLow  = bMedia + 2
		bFirstbyte = bResponse2		' low byte
		bSecondbyte = bResponse		' mid byte
		sWriteCommand				
		bLow  = bMedia + 6
		sWriteCommand
		bLow  = bMedia + 12
		sWriteCommand
		bLow  = bMedia + 16
		sWriteCommand
		
		bLow  = bMedia + 4			' Write the valid number of sectors (high byte)
		bFirstbyte = bTemp			' high byte
'		bSecondbyte = bTemp2		' highest byte (should be zero since largest possible sector count is 4,173,824 (DVD+DL))
		bSecondbyte = 0
		sWriteCommand				
		bLow  = bMedia + 8
		sWriteCommand
		bLow  = bMedia + 14
		sWriteCommand
		bLow  = bMedia + 18
		sWriteCommand
	#endif	

	bCmd  = 0xFF
	bHigh = 0x40
	bMid  = 0xBA
	bLow  = 0x06
	bNum  = 1
	bFirstbyte = 0
'	bSecondbyte = 0
WaitForF0:
	sWriteCommand							' Read login state
	if bResponse = 0xFF then goto WaitForF0	' In rare occations login state returns 0xFF and then we read it again
	if bResponse > 0xEF then 				' Wait for 0xF0 or above
		bCmd  = 0xFE
		bHigh = 0x00
		bMid  = 0x85
		bLow  = bMedia
'		bNum  = 1
		bFirstbyte = 0x38
'		bSecondbyte = 0
		sWriteCommand				' Write 0x38 to media (last step in backup patching)
		bBackup = 6					' Set backupflag, used by speed and region patching later
		set pLed1 off				' Patching is done, turn off LED1
		goto BackupPatchingDone		' Go to next step
	end if
	if bResponse < 0x51 then goto AllDone	' Avoid getting stuck in loop, exit if loginstate drops below 0x52
	goto WaitForF0							' Loop
		
BackupPatchingDone:		
	' Detect if disc is Wii or Gamecube
	bTemp = 0		' Set counter (256 loops)
	bTemp2 = 3		' Set counter2 (256 x 4 = 1024 loops)
	bCmd  = 0xFE	' Write bogus gametype byte
	bHigh = 0x40
	bMid  = 0x01
	bLow  = 0x80
'	bNum  = 1
	bFirstbyte = 0
'	bSecondbyte = 0
	sWriteCommand						
	bCmd  = 0xFF
'	bHigh = 0x40
'	bMid  = 0x01
'	bLow  = 0x80
	bNum  = 2
'	bFirstbyte = 0
'	bSecondbyte = 0
DetectGametype:
	sWriteCommand								' Read gametype byte
	if bResponse = 0x52 then goto DetectedWii	' 0x52 means Wii game (0x52 is Ascii for R)
	if bResponse = 0x47 then goto DetectedGC	' 0x47 means GC game (0x47 is Ascii for G)
	if bResponse = 0x50 then goto DetectedGC	' 0x50 means GC promotional game (0x50 is Ascii for P)
	if bResponse = 0x44 then goto DetectedGC	' 0x44 means GC demo game (0x44 is Ascii for D)
	#ifdef dConfigDisc
		if bResponse = 0x59 then				' 0x59 means yaosm config disc (0x59 is Ascii for Y)
			if bResponse2 = 0x41 then			' 0x41 means yaosm config disc (0x41 is Ascii for A)
				bGamecube = 2					' Set gamecube flag to 2 to indicate config disc
				#ifdef dDebug					' Flash both LEDs 3 times if config disc was detected
					repeat 3
						sDoubleLedFlash
					end repeat
				#endif
				goto AllDone					' No regionpatching for config disc so we end here
			end if
		end if
	#endif
	bTemp = bTemp - 1							' Decrease counter
	if bTemp = 0 then 
		if bTemp2 = 0 then
			bMyspeed = bMyspeed2				' Set drive to alternate Wii speed when we do a second start of the same disc
			goto AllDone						' Avoid getting stuck in loop (Second start of same disc exits here)
		end if
		bTemp2 = bTemp2 - 1						' Decrease counter2
	end if
	goto DetectGametype							' Loop

	' Patch Wii region
DetectedWii:
	bGamecube = 0				' Clear GC flag
	bMyspeed = bMyspeed1		' Set drive to Wii speed since this is the first start
	#ifdef dWiiRegionPatching 
		if bMyregion = 3 then goto WiiRegionPatchingDone					' Is regionpatching disabled?
		bTemp = 0															' Set counter (256 loops)
		bTemp2 = 1															' Set counter2 (256 x 2 = 512 loops)
		bCmd  = 0xFE														' Write bogus Wii region
		bHigh = 0xD1
		bMid  = 0xD3
		bLow  = 0xC9 + bBackup
		bNum  = 1
		bFirstbyte = 3
'		bSecondbyte = 0
		sWriteCommand
		bCmd  = 0xFF
		bFirstbyte = 0
DetectWiiRegion:
		sWriteCommand														' Read region
		if bResponse < 3 then												' Is it a valid region byte?
			if bResponse = bMyregion then goto WiiRegionPatchingDone		' Only patch if it is needed
			if bMyregion > 3 then goto SaveRegionSetting					' Autoconfigure region patching?
			set pLed1 on													' Turn on LED1 to show we are patching
			bCmd  = 0xFE													' Patch Wii region (WiiFree 2.37+ style)
			bFirstbyte = bMyregion
			sWriteCommand				
			bLow  = bRegionlow + bBackup									' Patch Wii region (WiiFree 2.37+ style)
			bFirstbyte = 0
			sWriteCommand								
			set pLed1 off													' Done patching, turn off LED1
			goto WiiRegionPatchingDone										' We're done here, go on to next step
		end if
		bTemp = bTemp - 1										' Decrease counter
		if bTemp = 0 then 
			if bTemp2 = 0 then goto WiiRegionPatchingDone		' Avoid getting stuck in loop
			bTemp2 = bTemp2 - 1									' Decrease counter2
		end if
		goto DetectWiiRegion
	#endif
		
WiiRegionPatchingDone:
	#ifdef dDebug				' Flash both LEDs 2 times if Wii game was detected
		repeat 2
			sDoubleLedFlash
		end repeat
	#endif
	goto AllDone				' Done here
	
DetectedGC:
	bGamecube = 1
	' Patch GC region
	#ifdef dGCRegionPatching
		if bMyregion = 3 then goto GCRegionPatchingDone				' Is regionpatching disabled?
		bTemp = 0													' Set counter (256 loops)
'		bCmd  = 0xFF
		bHigh = 0xD0
		bMid  = 0xF6
		bLow  = 0x61 + bBackup
		bNum  = 1
'		bFirstbyte = 0
'		bSecondbyte = 0
DetectGCRegion:
		sWriteCommand												' Read region
		if bResponse < 3 then										' Is it a valid region byte?
			if bResponse = bMyregion then goto GCRegionPatchingDone	' Only patch if it is needed
			if bMyregion > 3 then goto SaveRegionSetting			' Autoconfigure region patching?
			set pLed1 on											' Turn on LED1 to show we are patching
			bCmd = 0xFE												' Patch region
			bFirstbyte = bMyregion
			sWriteCommand											
			set pLed1 off											' Done patching, turn off LED1
			goto GCRegionPatchingDone								' We're done here, go on to next step
		end if
		bTemp = bTemp - 1											' Decrease counter
		if bTemp = 0 then goto GCRegionPatchingDone					' Avoid getting stuck in loop
		goto DetectGCRegion
	#endif
	
GCRegionPatchingDone:
	#ifdef dDebug				' Flash both LEDs 1 time if GC game was detected
		sDoubleLedFlash
	#endif
	goto AllDone				' Loop forever

SaveRegionSetting:
	EPWrite(0,bResponse)		' Save value in eprom pos 0
	#ifdef dEpromFlash
		sQuadrupleFlash			' Flash LED1 four times
	#endif
	goto ReadSettings			' Re-read configuration from eprom

'**************************** Sub routines *****************************************

' Sync with cn302
sub sSyncAndUnlock
ReSyncOnUnlock:
	' Sync
	bSyncbyte1 = 0			' Clear syncbyte1
	bSyncbyte2 = 0			' Clear syncbyte2
SyncLoop:
   	rotate bSyncbyte1 right	' Rotate syncbyte1 right
   	rotate bSyncbyte2 right	' Rotate syncbyte2 right
	set pClock on			' Clock high
	sClockDelay				' 4 us
	set pClock off			' Clock low
	sClockDelay				' 4 us
   	bSyncbyte1.7 = pInput	' Read bit a put it as highest bit in syncword
	set pClock on			' Clock high
	if bSyncbyte2 <> 0xEE then goto SyncLoop ' Loop until synced
	if bSyncbyte1 <> 0xEE then goto SyncLoop ' Loop until synced

Synced:
 	' Unlock after sync only if chip has been detected
 	if bUnlock > 0 then		
		' Write the unlock command (10 bytes)
		sWriteByte(0xFE)		' 0xff = READ, 0xfe = WRITE		
		sWriteByte(0)			' Always 0						
		sWriteByte(0x82)		' ADDR mid byte					
		sWriteByte(bUnlock)		' ADDR low byte					
		sWriteByte(0x41)		' first byte (write), 0x00 (read)	
		sWriteByte(0)			' ADDR high byte
		sWriteByte(0)			' second byte (write), 0x00 (read)	
		sWriteByte(0)			' Always 0						
		sWriteByte(1)			' bytes to write/read 1 or 2
		sWriteByte(0)			' Always 0					
		' Read the response
		sReadByte
		bUnlockStatus = bIObyte					' Byte 1 is 0x00 if OK
		sReadByte
		bUnlockStatus = bUnlockStatus + bIObyte	' Byte 2 is 0x00 if OK
		sReadByte
		bUnlockStatus = bUnlockStatus + bIObyte	' Byte 3 is first byte when we read, should be 0x41 in this case
		' Check if OK, if not, resync and try again
		if bUnlockStatus <> 0x41 then
			goto ReSyncOnUnlock		' resync
		end if
 	end if
end sub

sub sReadByte
	bIObyte = 0				' Clear inbyte
	repeat 8				' Read 8 bits (1 byte)
		rotate bIObyte right' Rotate inbyte right
		set pClock on		' Clock high
		sClockDelay			' 4 us
  		set pClock off		' Clock low
		sClockDelay			' 4 us
       	bIObyte.7 = pInput	' Read bit and put in as hightest bit in byte
		set pClock on		' Clock high
	end repeat
	sDelay					' Wait
end sub

sub sWriteByte(bIObyte) #NR
	repeat 8									' Write 8 bits (1 byte)
		set pClock on							' Clock high
		if bIObyte.0 = 0 then set pOutput off	' If LSB is 0 then set pOutput to low
		if bIObyte.0 = 1 then set pOutput on	' If LSB is 1 then set pOutput to high
		set pClock off							' Clock low
		sClockDelay								' 4 us
		set pClock on							' Clock high
		rotate bIObyte right					' Rotate to next bit
	end repeat
	sDelay										' Wait
end sub

sub sWriteCommand
	do
		' Write the command (10 bytes)
		sWriteByte(bCmd)						' 0xff = READ, 0xfe = WRITE		
		sWriteByte(0)							' Always 0						
		sWriteByte(bMid)						' ADDR mid byte					
		sWriteByte(bLow)						' ADDR low byte				
		sWriteByte(bFirstbyte)					' lowbyte (write), 0x00 (read)	
		sWriteByte(bHigh)						' ADDR high byte				
		sWriteByte(bSecondbyte)					' highbyte (write), 0x00 (read)	
		sWriteByte(0)							' Always 0						
		sWriteByte(bNum)						' bytes to write/read 1 or 2	
		sWriteByte(0)							' Always 0						
		' Read the response
		sReadByte
		bStatus = bIObyte						' Byte 1 is 0x00 if OK
		sReadByte
		bStatus = bStatus + bIObyte				' Byte 2 is 0x00 if OK
		sReadByte
		bResponse = bIObyte						' Byte 3 is first byte when we read
		if bNum = 2 then
	 		sReadByte
			bResponse2 = bIObyte				' Byte 4 is second byte when we read 2 bytes (1 word)
		end if
		if bStatus <> 0 then sSyncAndUnlock		' Check status_byte, if not OK, resync and try again
	loop until bStatus = 0 						' Exit if OK
end sub

sub sDelay
	wait 84 us
end sub

sub sClockDelay
end sub

#ifdef dEpromFlash
	sub sQuadrupleFlash
		repeat 4			' Flash LED1 four times to indicate that eprom has been updated
			wait 20 10ms
			set pLed1 on
			wait 20 10ms
			set pLed1 off
		end repeat
	end sub
#endif

#ifdef dDebug
	sub	sDoubleLedFlash
		wait 50 10ms
		set pLed1 on
		set pLed2 on
		wait 50 10ms
		set pLed1 off
		set pLed2 off
	end sub
#endif